# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Upload to cloud storage providers."""

from streamer.cloud.base import CloudUploaderBase


# Supported protocols.  Built based on which optional modules are available for
# cloud storage providers.
SUPPORTED_PROTOCOLS: list[str] = []


# All supported protocols.  Used to provide more useful error messages.
ALL_SUPPORTED_PROTOCOLS: list[str] = ['gs', 's3']


# Try to load the GCS (Google Cloud Storage) uploader.  If we can, the user has
# the libraries needed for GCS support.
try:
  from streamer.cloud.gcs import GCSUploader
  SUPPORTED_PROTOCOLS.append('gs')
except:
  pass


# Try to load the S3 (Amazon Cloud Storage) uploader.  If we can, the user has
# the libraries needed for S3 support.
try:
  from streamer.cloud.s3 import S3Uploader
  SUPPORTED_PROTOCOLS.append('s3')
except:
  pass


def create(upload_location: str) -> CloudUploaderBase:
  """Create an uploader appropriate to the upload location URL."""

  if upload_location.startswith("gs://"):
    return GCSUploader(upload_location)
  elif upload_location.startswith("s3://"):
    return S3Uploader(upload_location)
  else:
    raise RuntimeError("Protocol of {} isn't supported".format(upload_location))



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A pool of worker processes to upload to cloud storage."""

import abc
import enum
from setproctitle import setproctitle  # type: ignore
from queue import Queue

from typing import Optional
from typing_extensions import Self

import multiprocessing
# On Windows, we get multiprocessing.connection.PipeConnection.
# On Linux/macOS, we get multiprocessing.connection.Connection.
# Both inherit from multiprocessing.connection._ConnectionBase.
from multiprocessing.connection import _ConnectionBase

from streamer.cloud.base import CloudUploaderBase
import streamer.cloud.uploader as Uploader


class MessageType(enum.Enum):
  """Message type constants used for IPC from the main process to the pool."""

  WRITE_NON_CHUNKED = 'write_non_chunked'
  START_CHUNKED = 'start_chunked'
  WRITE_CHUNK = 'write_chunk'
  END_CHUNKED = 'end_chunked'
  DELETE = 'delete'
  RESET = 'reset'


class Message(object):
  """Message objects used for IPC from the main process to the pool."""
  def __init__(self, type: MessageType, path: str = "",
               data: bytes = b'') -> None:
    self.type: MessageType = type
    self.path: str = path
    self.data: bytes = data

  @staticmethod
  def write_non_chunked(path: str, data: bytes) -> 'Message':
    """A request to write non-chunked data (all at once)."""
    return Message(MessageType.WRITE_NON_CHUNKED, path, data)

  @staticmethod
  def start_chunked(path: str) -> 'Message':
    """A request to start a chunked data transfer."""
    return Message(MessageType.START_CHUNKED, path)

  @staticmethod
  def write_chunk(data: bytes) -> 'Message':
    """A request to write a single chunk of data."""
    return Message(MessageType.WRITE_CHUNK, data = data)

  @staticmethod
  def end_chunked() -> 'Message':
    """A request to end a chunked data transfer."""
    return Message(MessageType.END_CHUNKED)

  @staticmethod
  def delete(path: str) -> 'Message':
    """A request to delete a file."""
    return Message(MessageType.DELETE, path)

  @staticmethod
  def reset() -> 'Message':
    """A request to reset state when releasing a worker."""
    return Message(MessageType.RESET)


def worker_target(upload_location: str, reader: _ConnectionBase):
  """Target for multiprocessing.Process.

  This is the entry point for every worker subprocess.

  Reads messages from IPC and talks to cloud storage."""

  # Set the title of the process as it appears in "ps" under Linux.
  setproctitle('shaka-streamer cloud upload worker')

  # Create an uploader using whatever vendor-specific module is necessary for
  # this upload location URL.  (Google Cloud Storage, Amazon S3, etc.)
  uploader = Uploader.create(upload_location)

  # Wait for command messages from the main process, proxying each command to
  # the uploader.
  while True:
    try:
      message: Message = reader.recv()

      if message.type == MessageType.WRITE_NON_CHUNKED:
        uploader.write_non_chunked(message.path, message.data)
      elif message.type == MessageType.START_CHUNKED:
        uploader.start_chunked(message.path)
      elif message.type == MessageType.WRITE_CHUNK:
        uploader.write_chunk(message.data)
      elif message.type == MessageType.END_CHUNKED:
        uploader.end_chunked()
      elif message.type == MessageType.DELETE:
        uploader.delete(message.path)
      elif message.type == MessageType.RESET:
        uploader.reset()
    except EOFError:
      # Quit the process when the other end of the pipe is closed.
      return


class WorkerProcess(object):
  """A worker process and the write end of its pipe."""

  def __init__(self, process: multiprocessing.Process,
               writer: _ConnectionBase) -> None:
    self.process = process
    self.writer = writer


class AbstractPool(object):
  """An interface for a WorkerHandle (below) to talk to Pool (which references
  WorkerHandle).  Created to break a circular dependency for static typing."""

  @abc.abstractmethod
  def _release(self, process: WorkerProcess) -> None:
    """Add a process back into the pool."""
    pass


class WorkerHandle(CloudUploaderBase):
  """A proxy for a cloud uploader interface that sends commands to a worker
  process.  It is also a context manager for use with "with" statements."""

  def __init__(self, pool: AbstractPool, process: WorkerProcess) -> None:
    self._pool = pool
    self._process = process

  def __enter__(self) -> Self:
    # Part of the interface for context managers, but there's nothing to do
    # here.
    return self

  def __exit__(self, *args, **kwargs) -> None:
    """Reset the subprocess's uploader and release the subprocess back to the
    pool."""

    self._process.writer.send(Message.reset())
    self._pool._release(self._process)

  def write_non_chunked(self, path: str, data: bytes) -> None:
    self._process.writer.send(Message.write_non_chunked(path, data))

  def start_chunked(self, path: str) -> None:
    self._process.writer.send(Message.start_chunked(path))

  def write_chunk(self, data: bytes) -> None:
    self._process.writer.send(Message.write_chunk(data))

  def end_chunked(self) -> None:
    self._process.writer.send(Message.end_chunked())

  def delete(self, path: str) -> None:
    self._process.writer.send(Message.delete(path))

  def reset(self) -> None:
    # Part of the interface for uploaders, but this should not be called
    # explicitly.
    pass


class Pool(AbstractPool):
  """A pool of worker subprocesses that handle cloud upload actions."""

  def __init__(self, upload_location: str, size: int) -> None:
    self._all_processes: list[WorkerProcess] = []
    self._available_processes: Queue[WorkerProcess] = Queue()

    for i in range(size):
      reader, writer = multiprocessing.Pipe(duplex=False)
      process = multiprocessing.Process(target=worker_target,
                                        args=(upload_location, reader))
      process.start()
      worker_process = WorkerProcess(process, writer)
      self._available_processes.put(worker_process)
      self._all_processes.append(worker_process)

  def _release(self, worker_process: WorkerProcess) -> None:
    """Called by worker handles to release the worker back to the pool."""

    self._available_processes.put(worker_process)

  def get_worker(self) -> WorkerHandle:
    """Get an available worker.  Blocks until one is available.

    Returns a WorkerHandle meant to be used as a context manager (with "with"
    statements) so that it will be automatically released."""

    worker_process = self._available_processes.get(block=True)
    return WorkerHandle(self, worker_process)

  def close(self) -> None:
    """Close all worker processes."""

    for process in self._all_processes:
      process.writer.close()
      process.process.join()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Upload to Amazon S3."""

import urllib.parse

from typing import Any, Optional

import boto3  # type: ignore
import botocore.config  # type: ignore

from streamer.cloud.base import CloudUploaderBase


# S3 has a minimum chunk size for multipart uploads.
MIN_S3_CHUNK_SIZE = (5 << 20)  # 5MB


class S3Uploader(CloudUploaderBase):
  """See base class for interface docs."""

  def __init__(self, upload_location: str) -> None:
    # Parse the upload location (URL).
    url = urllib.parse.urlparse(upload_location)

    config = botocore.config.Config(retries = {'mode': 'standard'})
    self._client = boto3.client('s3', config=config)

    # If upload_location is "s3://foo/bar", url.netloc is "foo", which is the
    # bucket name.
    self._bucket_name = url.netloc

    # Strip both left and right slashes.  Otherwise, we get a blank folder name.
    self._base_path = url.path.strip('/')

    # State for chunked uploads:
    self._upload_id: Optional[str] = None
    self._upload_path: Optional[str] = None
    self._next_part_number: int = 0
    self._part_info: list[dict[str,Any]] = []
    self._data: bytes = b''

  def write_non_chunked(self, path: str, data: bytes) -> None:
    # No leading slashes, or we get a blank folder name.
    full_path = (self._base_path + path).strip('/')

    # Write the whole object at once.
    self._client.put_object(Body=data, Bucket=self._bucket_name, Key=full_path,
                            ExtraArgs={'CacheControl': 'no-cache'})

  def start_chunked(self, path: str) -> None:
    # No leading slashes, or we get a blank folder name.
    self._upload_path = (self._base_path + path).strip('/')

    # Ask the client to start a multi-part upload.
    response = self._client.create_multipart_upload(
        Bucket=self._bucket_name, Key=self._upload_path,
        CacheControl='no-cache')

    # This ID is sent to subsequent calls into the S3 client.
    self._upload_id = response['UploadId']

    # We must accumulate metadata about each part to complete the file at the
    # end of the chunked transfer.
    self._part_info = []
    # We must also number the parts.
    self._next_part_number = 1
    # Multi-part uploads for S3 can't have chunks smaller than 5MB.
    # We accumulate data for chunks here.
    self._data = b''

  def write_chunk(self, data: bytes, force: bool = False) -> None:
    # Collect data until we hit the minimum chunk size.
    self._data += data

    data_len = len(self._data)
    if data_len >= MIN_S3_CHUNK_SIZE or (data_len and force):
      # Upload one "part", which may be comprised of multiple HTTP chunks from
      # Packager.
      response = self._client.upload_part(
          Bucket=self._bucket_name, Key=self._upload_path,
          PartNumber=self._next_part_number, UploadId=self._upload_id,
          Body=self._data)

      # We have to collect this data, in this format, to finish the multipart
      # upload later.
      self._part_info.append({
        'PartNumber': self._next_part_number,
        'ETag': response['ETag'],
      })
      self._next_part_number += 1
      self._data = b''

  def end_chunked(self) -> None:
    # Flush the buffer.
    self.write_chunk(b'', force=True)

    # Complete the multipart upload.
    upload_info = { 'Parts': self._part_info }
    self._client.complete_multipart_upload(
        Bucket=self._bucket_name, Key=self._upload_path,
        UploadId=self._upload_id, MultipartUpload=upload_info)
    self.reset()

  def delete(self, path: str) -> None:
    self._client.delete_object(
        Bucket=self._bucket_name, Key=self._upload_path)

  def reset(self) -> None:
    self._upload_id = None
    self._upload_path = None
    self._next_part_number = 0
    self._part_info = []
    self._data = b''



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Upload to cloud storage providers.

Base class definition."""

import abc


class CloudUploaderBase(object):
  @abc.abstractmethod
  def write_non_chunked(self, path: str, data: bytes) -> None:
    """Write the non-chunked data to the destination."""
    pass

  @abc.abstractmethod
  def start_chunked(self, path: str) -> None:
    """Set up for a chunked transfer to the destination."""
    pass

  @abc.abstractmethod
  def write_chunk(self, data: bytes) -> None:
    """Handle a single chunk of data."""
    pass

  @abc.abstractmethod
  def end_chunked(self) -> None:
    """End the chunked transfer."""
    pass

  @abc.abstractmethod
  def delete(self, path: str) -> None:
    """Delete the file from cloud storage."""
    pass

  @abc.abstractmethod
  def reset(self) -> None:
    """Reset any chunked output state."""
    pass



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Upload to Google Cloud Storage."""

import urllib.parse

from typing import BinaryIO, Optional

import google.cloud.storage  # type: ignore
import google.api_core.exceptions  # type: ignore

from streamer.cloud.base import CloudUploaderBase


class GCSUploader(CloudUploaderBase):
  """See base class for interface docs."""

  def __init__(self, upload_location: str) -> None:
    # Parse the upload location (URL).
    url = urllib.parse.urlparse(upload_location)

    self._client = google.cloud.storage.Client()
    # If upload_location is "gs://foo/bar", url.netloc is "foo", which is the
    # bucket name.
    self._bucket = self._client.bucket(url.netloc)

    # Strip both left and right slashes.  Otherwise, we get a blank folder name.
    self._base_path = url.path.strip('/')

    # A file-like object from the Google Cloud Storage module that we write to
    # during a chunked upload.
    self._chunked_output: Optional[BinaryIO] = None

  def write_non_chunked(self, path: str, data: bytes) -> None:
    # No leading slashes, or we get a blank folder name.
    full_path = (self._base_path + path).strip('/')

    # An object representing the destination blob.
    blob = self._bucket.blob(full_path)
    blob.cache_control = 'no-cache'

    # A file-like interface to that blob.
    output = blob.open('wb', retry=google.cloud.storage.retry.DEFAULT_RETRY)
    output.write(data)
    output.close()

  def start_chunked(self, path: str) -> None:
    # No leading slashes, or we get a blank folder name.
    full_path = (self._base_path + path).strip('/')

    # An object representing the destination blob.
    blob = self._bucket.blob(full_path)
    blob.cache_control = 'no-cache'

    # A file-like interface to that blob.
    self._chunked_output = blob.open(
        'wb', retry=google.cloud.storage.retry.DEFAULT_RETRY)

  def write_chunk(self, data: bytes) -> None:
    assert self._chunked_output is not None
    self._chunked_output.write(data)

  def end_chunked(self) -> None:
    self.reset()

  def delete(self, path: str) -> None:
    # No leading slashes, or we get a blank folder name.
    full_path = (self._base_path + path).strip('/')
    blob = self._bucket.blob(full_path)
    try:
      blob.delete(retry=google.cloud.storage.retry.DEFAULT_RETRY)
    except google.api_core.exceptions.NotFound:
      # Some delete calls seem to throw "not found", but the files still get
      # deleted.  So ignore these and don't fail the request.
      pass

  def reset(self) -> None:
    if self._chunked_output:
      self._chunked_output.close()
      self._chunked_output = None
